"""Patch torch.onnx.export to use the exported program"""

from __future__ import annotations

import inspect
import io
import logging
import warnings
from typing import Any, Mapping, Sequence

import torch
import torch.export

from torch_onnx import _onnx_program, _core


logger = logging.getLogger(__name__)

WRITE_ERROR_REPORT = False
WRITE_PROFILE_REPORT = False


def _signature(model) -> inspect.Signature:
    should_be_callable = getattr(model, "forward", model)
    if callable(should_be_callable):
        return inspect.signature(should_be_callable)
    raise ValueError("model has no forward method and is not callable")


def _from_dynamic_axes_to_dynamic_shapes(
    model,
    dynamic_axes=None,
    input_names: Sequence[str] | None = None,
) -> dict[str, Any] | None:
    """

    dynamic_axes examples:
    (1) dynamic_axes = {"x": {0: "my_custom_axis_name_1"}, "y": {1: "my_custom_axis_name_2"}}
    (2) dynamic_axes = {"x": [0], "y": [1]}

    these will be converted to dynamic_shapes respectively:
    (1) dynamic_shapes = {"x": {0: Dim("my_custom_axis_name_1")}, "y": {1: Dim("my_custom_axis_name_2")}}
    (2) dynamic_shapes = {"x": {0: Dim("x_dim_0")}, "y": {1: Dim("y_dim_1")}}  # auto-generated dim names

    """
    # https://github.com/pytorch/pytorch/pull/128371
    if dynamic_axes is None:
        return None

    input_names_set = set() if input_names is None else set(input_names)

    dynamic_shapes = {}
    for input_name, axes in dynamic_axes.items():
        if input_name in input_names_set:
            raise ValueError(
                "input names is not supported yet. Please use model forward signature."
            )
        if isinstance(axes, dict):
            dynamic_shapes[input_name] = {
                k: torch.export.Dim(v) for k, v in axes.items()
            }
        elif isinstance(axes, list):
            dynamic_shapes[input_name] = {
                k: torch.export.Dim(f"{input_name}_dim_{k}") for k in axes
            }
        else:
            raise TypeError(
                f"dynamic_axes value must be either a dict or a list, but got {type(axes)}"
            )
    # torch.export.export needs static dim to present in dynamic_shapes
    # for all input tensors, so we need to add them with None
    try:
        sig = _signature(model)
    except ValueError as e:
        warnings.warn(
            f"{e}, skipping auto filling None on static axes...", stacklevel=1
        )
        return dynamic_shapes
    for input_name in sig.parameters:
        if input_name not in dynamic_shapes:
            dynamic_shapes[input_name] = None
    return dynamic_shapes


def _get_torch_export_args(
    model: torch.nn.Module | torch.export.ExportedProgram,
    args: tuple[Any, ...],
    kwargs: dict[str, Any] | None,
    dynamic_axes: Mapping[str, Mapping[int, str]] | Mapping[str, Sequence[int]] | None,
    input_names: Sequence[str] | None,
):
    if not kwargs and args and isinstance(args[-1], dict):
        kwargs = args[-1]
        args = args[:-1]

    dynamic_shapes = _from_dynamic_axes_to_dynamic_shapes(
        model, dynamic_axes, input_names
    )
    return args, kwargs, dynamic_shapes


def _torch_onnx_export(
    model: torch.nn.Module | torch.export.ExportedProgram,
    args: tuple[Any, ...],
    f: str | io.BytesIO | None = None,
    *,
    kwargs: dict[str, Any] | None = None,
    export_params: bool = True,
    input_names: Sequence[str] | None = None,
    output_names: Sequence[str] | None = None,
    opset_version: int | None = None,
    dynamic_axes: Mapping[str, Mapping[int, str]]
    | Mapping[str, Sequence[int]]
    | None = None,
    profile: bool = False,
    error_report: bool = False,
    **_,
) -> _onnx_program.ONNXProgram:
    # Set up the error reporting facilities
    error_report = WRITE_ERROR_REPORT or error_report
    profile = WRITE_PROFILE_REPORT or profile

    if not isinstance(model, torch.export.ExportedProgram):
        args, kwargs, dynamic_shapes = _get_torch_export_args(
            model, args, kwargs, dynamic_axes, input_names
        )
    else:
        args, kwargs, dynamic_shapes = _get_torch_export_args(
            model, args, kwargs, None, input_names
        )

    onnx_program = _core.export(
        model,
        args,
        kwargs,
        registry=None,
        dynamic_shapes=dynamic_shapes,
        input_names=input_names,
        output_names=output_names,
        profile=profile,
        error_report=error_report,
    )

    if f is not None:
        onnx_program.save(f, include_initializers=export_params)

    return onnx_program


def _torch_onnx_dynamo_export(
    model: torch.nn.Module | torch.export.ExportedProgram,
    /,
    *model_args,
    export_options: torch.onnx.ExportOptions | None = None,
    **model_kwargs,
) -> _onnx_program.ONNXProgram:
    if export_options and export_options.dynamic_shapes:
        warnings.warn("Dynamic shapes are not implemented yet.", stacklevel=1)

    return _core.export(
        model,
        model_args,
        kwargs=model_kwargs,
    )


_original_torch_onnx_export = torch.onnx.export
_original_torch_onnx_utils_export = torch.onnx.utils._export
_original_torch_onnx_dynamo_export = torch.onnx.dynamo_export


def patch_torch(error_report: bool = False, profile: bool = False):
    global WRITE_ERROR_REPORT  # noqa: PLW0603
    WRITE_ERROR_REPORT = error_report
    global WRITE_PROFILE_REPORT  # noqa: PLW0603
    WRITE_PROFILE_REPORT = profile
    torch.onnx.export = _torch_onnx_export
    torch.onnx.utils._export = _torch_onnx_export
    torch.onnx.dynamo_export = _torch_onnx_dynamo_export


def unpatch_torch():
    torch.onnx.export = _original_torch_onnx_export
    torch.onnx.utils._export = _original_torch_onnx_utils_export
    torch.onnx.dynamo_export = _original_torch_onnx_dynamo_export



import torch

from torch_onnx import _analysis, _registration
from onnxscript import ir


def _format_export_status(step: int, error: bool):
    def status_emoji(current_step: int, error: bool) -> str:
        if current_step < step:
            return "✅"
        if current_step == step:
            return "❌" if error else "✅"
        return "⚪"

    return (
        f"```\n"
        f"{status_emoji(0, error)} Obtain model graph with `torch.export.export`\n"
        f"{status_emoji(1, error)} Translate the graph into ONNX\n"
        f"{status_emoji(2, error)} Run `onnx.checker` on the ONNX model\n"
        f"{status_emoji(3, error)} Execute the model with ONNX Runtime\n"
        f"{status_emoji(4, error)} Validate model output accuracy\n"
        f"```\n\n"
    )


def create_torch_export_error_report(
    filename: str, formatted_traceback: str, *, profile_result: str | None
):
    with open(filename, "w", encoding="utf-8") as f:
        f.write("# PyTorch ONNX Conversion Error Report\n\n")
        f.write(_format_export_status(0, True))
        f.write("Error message:\n\n")
        f.write("```pytb\n")
        f.write(formatted_traceback)
        f.write("```\n\n")
        if profile_result is not None:
            f.write("## Profiling result\n\n")
            f.write("```\n")
            f.write(profile_result)
            f.write("```\n")


def create_onnx_export_error_report(
    filename: str,
    formatted_traceback: str,
    program: torch.export.ExportedProgram,
    *,
    step: int,
    profile_result: str | None,
    model: ir.Model | None = None,
    registry: _registration.OnnxRegistry | None = None,
):
    with open(filename, "w", encoding="utf-8") as f:
        f.write("# PyTorch ONNX Conversion Error Report\n\n")
        f.write(_format_export_status(step, True))
        f.write("Error message:\n\n")
        f.write("```pytb\n")
        f.write(formatted_traceback)
        f.write("```\n\n")
        f.write("Exported program:\n\n")
        f.write("```python\n")
        f.write(str(program))
        f.write("```\n\n")
        if model is not None:
            f.write("ONNX model:\n\n")
            f.write("```python\n")
            f.write(str(model))
            f.write("\n```\n\n")
        f.write("## Analysis\n\n")
        _analysis.analyze(program, file=f, registry=registry)
        if profile_result is not None:
            f.write("\n## Profiling result\n\n")
            f.write("```\n")
            f.write(profile_result)
            f.write("```\n")


def crete_onnx_export_profile_report(
    filename: str, program: torch.export.ExportedProgram, profile_result: str, step: int
):
    """Create a report for the ONNX export profiling result.

    Args:
        filename: The file to write the report to.
        program: The exported program.
        profile_result: The profiling result.
        step: The current step in the conversion process.
    """
    with open(filename, "w", encoding="utf-8") as f:
        f.write("# PyTorch ONNX Conversion Report\n\n")
        f.write(_format_export_status(step, False))
        f.write("Exported program:\n\n")
        f.write("```python\n")
        f.write(str(program))
        f.write("```\n\n")
        f.write("## Profiling result\n\n")
        f.write("```\n")
        f.write(profile_result)
        f.write("```\n")



import logging
import os
import pathlib
from typing import IO

import onnx
import torch
from onnxscript import ir

logger = logging.getLogger(__name__)


class ONNXProgram:
    """A substitute class for `torch.onnx.ONNXProgram`."""

    def __init__(self, model: ir.Model, exported_program: torch.export.ExportedProgram):
        self.model = model
        self.exported_program = exported_program

    @property
    def model_proto(self) -> onnx.ModelProto:
        """Compatibility property for `torch.onnx.ONNXProgram.model_proto`."""
        return ir.serde.serialize_model(self.model)

    def save(
        self,
        destination: str | os.PathLike | IO[bytes],
        *,
        include_initializers: bool = True,
        external_data: bool | None = None,
        **_,
    ):
        """Save the ONNX model to the specified destination.

        When `external_data` is `True` or the model is larger than 2GB,
        the weights are saved as external data in a separate file.

        Args:
            destination: The path to save the ONNX model to.
            include_initializers: Whether to include the initializers in the saved model.
            external_data: Whether to save the weights as external data in a separate file.

        Raises:
            TypeError: If `external_data` is `True` and `destination` is not a file path.
        """
        if not include_initializers:
            self.model.graph.initializers.clear()
            logger.warning(
                "The initializers have been removed from the model. This is destructive. "
                "Developers: Please implement ir.Model copy() and remove initializers on the copied model."
            )
        proto = ir.serde.serialize_model(self.model)
        byte_size = proto.ByteSize()
        model_too_large = (byte_size) >= 1 << 31
        if external_data or model_too_large:
            # TODO: Create an IR pass to handle external tensors conversion
            if model_too_large:
                logger.warning(
                    "The serialized ONNX model is larger than 2GB (%s). "
                    "Saving the weights as external data in a separate file.",
                    byte_size,
                )
            if not isinstance(destination, (str, os.PathLike)):
                raise TypeError(
                    "Saving the weights as external data is only supported when destination is a file path"
                )
            destination_path = pathlib.Path(destination)
            data_path = destination_path.with_suffix(f"{destination_path.suffix}.data")
            onnx.save_model(
                proto,
                destination,
                save_as_external_data=True,
                location=os.fspath(data_path),
            )
        else:
            onnx.save_model(proto, destination)



from __future__ import annotations

from typing import Sequence

from onnxscript import ir


def rename_inputs(model: ir.Model, new_names: Sequence[str]):
    # TODO: Ensure the names do not have duplicates
    for input, new_name in zip(model.graph.inputs, new_names):
        input.name = new_name


def rename_outputs(model: ir.Model, new_names: Sequence[str]):
    for output, new_name in zip(model.graph.outputs, new_names):
        output.name = new_name



from __future__ import annotations

import io
import torch.export


def verify_model(
    exported_program: torch.export.ExportedProgram,
    onnx_model: str | io.BytesIO,
    args,
    kwargs,
):
    # TODO: Implement
    pass
